// XMLMimeFilterPP.cpp : Implementation of CXMLMimeFilterPP
#include "stdafx.h"
#include "Xmlmimefilter.h"
#include "XMLMimeFilterPP.h"

static TCHAR TastyHtmlEnd[] = _T("\r\n</BODY></HTML>\r\n");

// If we are sending UNICODE data back, we need to add a charset label for MSHTML
// to understand that it is not the default, ANSI text
#ifdef UNICODE

#define CONTENT_TYPE L"text/html"
static TCHAR TastyHtmlStart[] =
_T("<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; CHARSET=unicode\">\r\n<HTML>\r\n<HEAD>\r\n<TITLE>barbee's XML Source View</TITLE>\r\n</HEAD>\r\n<BODY>\r\n");

#else !UNICODE

#define CONTENT_TYPE L"text/html"
static TCHAR TastyHtmlStart[] =
_T("<HTML>\r\n<HEAD>\r\n<TITLE>barbee's XML Source View</TITLE>\r\n</HEAD>\r\n<BODY>\r\n");

#endif

#define BUFFER_READ_SIZE 2048
static BYTE g_buffer[BUFFER_READ_SIZE + sizeof(TCHAR)];

/////////////////////////////////////////////////////////////////////////////
// CXMLMimeFilterPP

CXMLMimeFilterPP::CXMLMimeFilterPP()
{
	m_pHTMLBuffer = NULL;
	m_bFirstRead = true;
	m_pOutgoingProtSink = NULL;
	m_pIncomingProt = NULL;
	m_grfSTI = 0;
	m_bIncomingIsUnicode = false;
	m_bSplitUniWord = false;
}

CXMLMimeFilterPP::~CXMLMimeFilterPP()
{
	delete m_pHTMLBuffer;

	if (m_pIncomingProt)
		m_pIncomingProt->Release();
}

/////////////////////////////////////////////////////////////////////////////
// IInternetProtocol interface

STDMETHODIMP CXMLMimeFilterPP::Start(
        LPCWSTR szUrl,
        IInternetProtocolSink *pIProtSink,
        IInternetBindInfo *pIBindInfo,
        DWORD grfSTI,
        DWORD dwReserved)
{
USES_CONVERSION;	
ATLTRACE(_T("Start - url=%s, sti=%x\n"),OLE2T((LPWSTR)szUrl), grfSTI);

	HRESULT hr = E_FAIL;
	m_grfSTI      = grfSTI;
	m_bIncomingIsUnicode = false;
	m_bSplitUniWord = false;

	if (!(grfSTI & PI_FILTER_MODE))
    {
        hr = E_INVALIDARG;
    }
    else
    {
		// get the protocol pointer from reserved pointer
        PROTOCOLFILTERDATA* ProtFiltData = (PROTOCOLFILTERDATA*) dwReserved;
        _ASSERTE(NULL == m_pIncomingProt);
		if (NULL == ProtFiltData->pProtocol)
		{
			// !! We can't do anything without an interface to read from
			_ASSERTE(false);
			return E_INVALIDARG;
		}

        m_pIncomingProt = ProtFiltData->pProtocol;
        m_pIncomingProt->AddRef();

        // hold onto the sink as well
        _ASSERTE(NULL == m_pOutgoingProtSink);
        m_pOutgoingProtSink = pIProtSink;
        m_pOutgoingProtSink->AddRef();

		m_bFirstRead = true;

		m_pHTMLBuffer = new CStringDataBuffer(2048);
        
        hr = S_OK;
    }

	return hr;
}

STDMETHODIMP CXMLMimeFilterPP::Continue(PROTOCOLDATA *pStateInfo)
{
ATLTRACE(_T("Continue\n"));

	if (NULL == m_pIncomingProt)
		return E_FAIL;

	return m_pIncomingProt->Continue(pStateInfo);
}

STDMETHODIMP CXMLMimeFilterPP::Abort(HRESULT hrReason,DWORD dwOptions)
{
ATLTRACE(_T("Abort - reason=%x\n"), hrReason);

	if (NULL == m_pIncomingProt)
		return E_FAIL;

	return m_pIncomingProt->Abort(hrReason, dwOptions);
}

STDMETHODIMP CXMLMimeFilterPP::Terminate(DWORD dwOptions)
{
ATLTRACE(_T("Terminate - options=%x\n"), dwOptions);

	// Release the sink
	if (m_pOutgoingProtSink)
		m_pOutgoingProtSink->Release();

	if (NULL == m_pIncomingProt)
		return E_FAIL;

	return m_pIncomingProt->Terminate(dwOptions);
}
STDMETHODIMP CXMLMimeFilterPP::Suspend()
{
ATLTRACENOTIMPL("Suspend\n");

	if (NULL == m_pIncomingProt)
		return E_FAIL;

	return m_pIncomingProt->Suspend();
}

STDMETHODIMP CXMLMimeFilterPP::Resume()
{
ATLTRACENOTIMPL("Resume\n");
	
	if (NULL == m_pIncomingProt)
		return E_FAIL;

	return m_pIncomingProt->Resume();
}

STDMETHODIMP CXMLMimeFilterPP::Read(void *pv, ULONG cb, ULONG *pcbRead)
{
	ATLTRACE(_T("IIP::Read - pv = 0x%x, cb = %d\n"), pv, cb);
	HRESULT hr = S_OK;
	HRESULT hrFromProt;

	char* pBuff = (char*)malloc(cb);

	//hrFromProt = m_pIncomingProt->Read( pv, cb, pcbRead );
	hrFromProt = m_pIncomingProt->Read( (void*)pBuff, cb, pcbRead );

	if (*pcbRead > 0)
	{
		char* pCur = pBuff;
		char* pEnd = pBuff + *pcbRead;

		while (pCur < pEnd)
		{
			if (*pCur == '')
				*pCur = '';

			pCur++;
		}

		memcpy( pv, (void*)pBuff, *pcbRead);
	}

	free( pBuff );

	return hrFromProt;

/*
	BYTE* pbuff = g_buffer;
	ULONG cbReadFromProt; 
	DWORD dwIsUnicode;
	ULONG cbUniCheck;

	if (!m_pHTMLBuffer->IsDone())
	{
		// If there was a split unicode word on the last read from the prot handler, fix it by
		// sticking the last byte onto the front of this read
		if (m_bIncomingIsUnicode && m_bSplitUniWord)
		{
			pbuff[0] = m_byteSplit;
			pbuff++;
			m_bSplitUniWord = false;
		}
		// Retrieve Requested Amount of Data from Protocol Handler
		hrFromProt = m_pIncomingProt->Read((void*)pbuff, BUFFER_READ_SIZE, &cbReadFromProt);
		if (E_PENDING != hrFromProt) // We may have some extra data in our buffer, so continue
		{
			if (FAILED(hrFromProt))
				return hrFromProt;
			else 
			{
				// Parse Data into String Data Buffer
				if (cbReadFromProt > 0)
				{
					if (m_bFirstRead)
					{

						if (cbReadFromProt >= 2 && pbuff[0] == 0xFF && pbuff[1] == 0xFE)
						{
							m_bIncomingIsUnicode = true;
							pbuff += 2;
						}
						else
						{
							// even # of bytes read from prot handler?
							if (cbReadFromProt & 0x1)
								cbUniCheck = cbReadFromProt - 1;
							else
								cbUniCheck = cbReadFromProt;
												
							dwIsUnicode = IsTextUnicode(pbuff, cbUniCheck, NULL);
							if (0 != dwIsUnicode)
								m_bIncomingIsUnicode = true;
						}

						BeginXMLConversion();
						m_bFirstRead = false;
					}
					hr = ParseXMLIntoHTML(pbuff, cbReadFromProt);
					if (FAILED(hr))
						return hr;
				}
				
				// Prot Handler will let us know when there is no more data left
				if (S_FALSE == hrFromProt)
				{
					EndXMLConversion();
					m_pHTMLBuffer->Done();
				}
			}
		}
	}
    // this filter converts text/xml to text/html so we switch the MIME type here
	m_pOutgoingProtSink->ReportProgress(BINDSTATUS_VERIFIEDMIMETYPEAVAILABLE, CONTENT_TYPE);

	// Read Requested Amount of Data from String Data Buffer (there probably will be leftover)
	// & write it to pv
	// String buffer is expected to return S_FALSE when we hit the end
	hr = m_pHTMLBuffer->ReadData(pv, cb, pcbRead); 
	ATLTRACE(_T("...pcbRead = %d\n"), *pcbRead);
	
	return hr;
*/
}

STDMETHODIMP CXMLMimeFilterPP::Seek(
        LARGE_INTEGER dlibMove,
        DWORD dwOrigin,
        ULARGE_INTEGER *plibNewPosition)
{
ATLTRACENOTIMPL("Seek\n");
	
	return E_NOTIMPL;
}

STDMETHODIMP CXMLMimeFilterPP::LockRequest(DWORD dwOptions)
{
ATLTRACE(_T("LockRequest\n"));

	return S_OK;
}

STDMETHODIMP CXMLMimeFilterPP::UnlockRequest()
{
ATLTRACE(_T("UnlockRequest\n"));

	return S_OK;
}

void CXMLMimeFilterPP::BeginXMLConversion()
{
	*m_pHTMLBuffer << TastyHtmlStart; // Writes HTML, BODY start tag
}

void CXMLMimeFilterPP::EndXMLConversion()
{
	*m_pHTMLBuffer << TastyHtmlEnd; // Writes BODY, HTML end tags
}

// pInbuff points to a buffer of bytes, not characters, and should always have 1 extra character's worth
HRESULT CXMLMimeFilterPP::ParseXMLIntoHTML(BYTE* pInBuff, ULONG cbBufferSize)
{
ATLTRACE(_T("ParseXMLIntoHTML\n"));
	USES_CONVERSION;
	HRESULT hr = E_FAIL;

	if (NULL == pInBuff)
	{
		_ASSERTE(false);
		return E_INVALIDARG;
	}
	
	TCHAR* pbuff = (TCHAR*)pInBuff;
	ULONG cchBuffer = cbBufferSize;

	// Conversion from whatever is in buffer (we don't know if it is uni or not) to TCHAR

	// TODO: Obviously our conversion of data here is very simplistic - coding an improvement
	// over A2T and W2T conversion routines of the data would be a nice exercise for the 
	// exuberant MIME filter developer

	if (m_bIncomingIsUnicode)
	{
		// Prot Handler split last unicode word in middle
		if (cbBufferSize & 0x1)
		{
			m_bSplitUniWord = true;
			m_byteSplit = pInBuff[cbBufferSize-1];
			cchBuffer--;
			pInBuff[cbBufferSize-1] = 0;
			pInBuff[cbBufferSize] = 0;
		}
		else
		{
			pInBuff[cbBufferSize] = 0;
			pInBuff[cbBufferSize+1] = 0;
		}

		cchBuffer = cchBuffer / sizeof(TCHAR);
		pbuff = W2T((LPWSTR)pbuff);

	}
	else
	{
		pInBuff[cbBufferSize] = 0;
		pbuff = A2T((LPSTR)pbuff);
	}

	// TODO: We should also do some MIME sniffing tests to verify that what we
	//		 actually have is text data.. or look for an XML DTD PI or something like that

	bool bDone = false;
	TCHAR* pCur = pbuff;
	TCHAR* pNewCur = NULL;

	_ASSERTE(m_pHTMLBuffer);
	
	TCHAR* strCharSet = _T("<>&\r");

	bDone = false;
	TCHAR* pEnd = pbuff + cchBuffer;
	while (!bDone && (pCur < pEnd))
	{
		pNewCur = _tcspbrk(pCur, strCharSet);

		if (NULL == pNewCur)
		{
			m_pHTMLBuffer->WriteString(pCur, pEnd - pCur);
			bDone = true;
		}
		else
		{
			// Write out part up to <>&..
			if (pNewCur > pCur)
				m_pHTMLBuffer->WriteString(pCur, pNewCur - pCur); // no Null Terminator
			switch (*pNewCur)
			{
			case _T('<'):
				*m_pHTMLBuffer << _T("&lt;");
				break;

			case _T('>'):
				*m_pHTMLBuffer << _T("&gt;");
				break;

			case _T('&'):
				*m_pHTMLBuffer << _T("&amp;");
				break;

			case _T('\r'):
				*m_pHTMLBuffer << _T("\r<BR>");
				break;

			default:
				_ASSERTE(false); // should never hit this!
				break;
			}

			if (0 == *pCur)
				bDone = true;

			pCur = pNewCur + 1;
		}		
	}

	return S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// IInternetProtocolSink interface
STDMETHODIMP CXMLMimeFilterPP::Switch(PROTOCOLDATA __RPC_FAR *pProtocolData)
{
	ATLTRACE(_T("IIPS::Switch\n"));
	HRESULT hr = S_OK; 

	if (NULL == m_pOutgoingProtSink)
		return E_FAIL;

	return m_pOutgoingProtSink->Switch(pProtocolData);
}

STDMETHODIMP CXMLMimeFilterPP::ReportProgress(ULONG ulStatusCode, LPCWSTR szStatusText)
{
	USES_CONVERSION;
	ATLTRACE(_T("IIPS::ReportProgress - status = (%d) %s\n"), ulStatusCode, W2CT(szStatusText));
	HRESULT hr = S_OK; 

	if (NULL == m_pOutgoingProtSink)
		return E_FAIL;

	return m_pOutgoingProtSink->ReportProgress(ulStatusCode, szStatusText);
}

STDMETHODIMP CXMLMimeFilterPP::ReportData(DWORD grfBSCF, ULONG ulProgress, ULONG ulProgressMax)
{
	ATLTRACE(_T("IIPS::ReportData - bscf = %d, prog = %d, max = %d\n"),
					grfBSCF, ulProgress, ulProgressMax);
	HRESULT hr = S_OK; 

	if (NULL == m_pOutgoingProtSink)
		return E_FAIL;

	return m_pOutgoingProtSink->ReportData(grfBSCF, ulProgress, ulProgressMax);
}

STDMETHODIMP CXMLMimeFilterPP::ReportResult(HRESULT hrResult, DWORD dwError, LPCWSTR szResult)
{
	USES_CONVERSION;
	ATLTRACE(_T("IIPS::ReportResult - result = %x, error = (%d) %s\n"),
					hrResult, dwError, W2CT(szResult));
	HRESULT hr = S_OK; 

	if (NULL == m_pOutgoingProtSink)
		return E_FAIL;

	return m_pOutgoingProtSink->ReportResult(hrResult, dwError, szResult);
}